﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;
using System.Threading;
using System.Diagnostics;
#if MEMORY_MAPPED_FILES_DOTNET4
using System.IO.MemoryMappedFiles;
#endif

using FileMapped;

namespace FileMappedImplementations
{
    /// <summary>
    /// Memory mapped conversion functions. This class is unsafe as it uses fixed pointers.
    /// </summary>
    public unsafe class InteropMemoryMappedCciImplementation
    {
        /// <summary>
        /// Perform multithreaded Mmeory Mapped conversion.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns>Number of bytes processed</returns>
        public Int64 cciMMMT(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            Int64 iPos = 0;
            byte* pIn = null, pOut = null;
            Int64 totalBytesEncrypted = 0;
            FileMm inMap = null, outMap = null;

            try
            {
                FileInfo fi = new FileInfo(inFile);
                Int64 fileSize = fi.Length;

                inMap = new FileMm(inFile);
                outMap = new FileMm(outFile, false, fileSize);

                try
                {
                    pIn = inMap.pFile;
                    pOut = outMap.pFile;

                    // Create the conversion thread arguments, then create the threads to convert file stripes concurrently
                    cciArgMM[] threadArgs = (from n in Enumerable.Range(0, numberStripes) select new cciArgMM(pIn, pOut, iShift, n, fileLength, numberStripes, bufferSize, optimize)).ToArray();
                    Thread[] cciThread = (from n in Enumerable.Range(0, numberStripes) select new Thread(new ParameterizedThreadStart(cciStripeMM))).ToArray();

                    // Start all the threads. The static ForAll method is implemented at the end of the file.
                    threadArgs.ForAll((ta) => cciThread[ta.threadNumber].Start(ta));

                    // Join the threads, accumulate the byte count
                    threadArgs.ForAll((ta) => { cciThread[ta.threadNumber].Join(); totalBytesEncrypted += ta.numBytesEncrypted; });
                }

                catch (Exception e)
                {
                    Console.Error.WriteLine("Error. cciMMMT. Bytes encrypted: {0}. {1}.", iPos, e.Message);
                    return iPos;
                }
            }
            finally
            {
                inMap.Dispose();
                outMap.Dispose();
            }
            return totalBytesEncrypted;
        }

        /// <summary>
        /// Perform Mmeory Mapped conversion with multiple work items in a thread pool
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns>Number of bytes processed</returns>
        public Int64 cciMMTP(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            Int64 iPos = 0;
            byte* pIn = null, pOut = null;
            FileMm inMap = null, outMap = null;
            Int64 totalBytesEncrypted = 0;

            try
            {
                FileInfo fi = new FileInfo(inFile);
                Int64 fileSize = fi.Length;

                inMap = new FileMm(inFile);
                outMap = new FileMm(outFile, false, fileSize);
                try
                {
                    pIn = inMap.pFile;
                    pOut = outMap.pFile;

                    Thread[] cciThread = new Thread[numberStripes];

                    int remainingCallbacks = numberStripes;
                    // Create the conversion work items to convert file stripes concurrently.
                    ManualResetEvent finishedEvent = new ManualResetEvent(false);

                    // Create the conversion thread/task arguments
                    cciArgMM[] threadArgs = (from n in Enumerable.Range(0, numberStripes) select new cciArgMM(pIn, pOut, iShift, n, fileLength, numberStripes, bufferSize, optimize)).ToArray();

                    // Create the pooled tasks and wait for them all to complete.
                    threadArgs.ForAll((ta) =>
                        ThreadPool.QueueUserWorkItem(delegate
                        {
                            cciStripeMM(ta);
                            if (Interlocked.Decrement(ref remainingCallbacks) == 0)
                            {
                                finishedEvent.Set();
                            }
                        })
                    );

                    // Wait for all pooled tasks to complete
                    finishedEvent.WaitOne();

                    // Accumulate the encrypted byte count
                    threadArgs.ForAll((ta) => { totalBytesEncrypted += ta.numBytesEncrypted; });
                }

                catch (Exception e)
                {
                    Console.Error.WriteLine("Error. cciMMTP. Bytes encrypted: {0}. {1}.", iPos, e.Message);
                    return iPos;
                }
            }
            finally
            {
                inMap.Dispose();
                outMap.Dispose();
            }
            return totalBytesEncrypted;
        }

        /// <summary>
        /// cciStripeMM thread/work item function. MemoryMapped version.
        /// </summary>
        /// <param name="cargObject"></param>
        /// <remarks>This function contains both optimized and non-optimized code. The code to use is determined at run time based on the thread parameter.
        /// The optimized code is a "first attempt", and you can experiment with alternative implementations.</remarks>
        public void cciStripeMM(object cciArgMMObject)
        {
            cciArgMM cciArgMM = cciArgMMObject as cciArgMM;
            if (cciArgMM == null)
            {
                // This should never happen.
                Console.Error.WriteLine("Error. cciStripeMM thread function. Argument object is not a cciArgMM.");
                return; // The error will be detected as a failure to convert all the bytes in the file.
            }

            int iThread = cciArgMM.threadNumber;
            byte iShift = (byte)cciArgMM.iShift;
            int numberStripes = cciArgMM.numberStripes;
            int bufferSize = cciArgMM.bufferSize;
            Int64 fileLength = cciArgMM.fileLength, numBytesEncrypted = cciArgMM.numBytesEncrypted;
            bool optimize = cciArgMM.optimize;

            Int64 iPos = 0, iStripe = iThread;

            byte* pInBase = cciArgMM.pInBase, pOutBase = cciArgMM.pOutBase, pIn = null, pOut = null;

            try
            {
                while (iStripe * bufferSize < fileLength)
                {
                    pIn = pInBase + (iStripe * bufferSize);
                    pOut = pOutBase + (iStripe * bufferSize);
                    Int64 maxPosition = Math.Min(bufferSize, fileLength - iStripe * bufferSize);

                    byte b;
                    if (!optimize)
                    {
                        for (iPos = 0; iPos < maxPosition; iPos++)
                        {
                            b = *pIn;
                            *pOut = (byte)(b + iShift);
                            pIn++; pOut++;
                        }
                    }
                    else
                    {
                        for (iPos = 0; iPos < maxPosition; iPos += 8)
                        {
                            *pOut = (byte)(*pIn + iShift);
                            *(pOut + 1) = (byte)(*(pIn + 1) + iShift);
                            *(pOut + 2) = (byte)(*(pIn + 2) + iShift);
                            *(pOut + 3) = (byte)(*(pIn + 3) + iShift);
                            *(pOut + 4) = (byte)(*(pIn + 4) + iShift);
                            *(pOut + 5) = (byte)(*(pIn + 5) + iShift);
                            *(pOut + 6) = (byte)(*(pIn + 6) + iShift);
                            *(pOut + 7) = (byte)(*(pIn + 7) + iShift);
                            pIn += 8; pOut += 8;
                        }
                        pIn -= 8; pOut -= 8; iPos -= 8;
                        for (; iPos < maxPosition; iPos++)
                        {
                            b = *pIn;
                            *pOut = (byte)(b + iShift);
                            pIn++; pOut++;
                        }
                    }

                    numBytesEncrypted += maxPosition;
                    iStripe += numberStripes;
                }
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error. cciStripeMM read/write loop thread {0} at position {1} in stripe {2}. {3}", iThread, iPos, iStripe, e.Message);
            }

            cciArgMM.numBytesEncrypted = numBytesEncrypted;
            return;
        }

        /// <summary>
        /// Thread argument for the cci threads and work items. MemoryMapped version.
        /// </summary>
        class cciArgMM
        {
            private Int64 _numBytesEncrypted = 0;

            public cciArgMM(byte* pIn, byte* pOut, int iShift, int iThread, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
            {
                _threadNum = iThread;
                _iShift = iShift;
                _fileLength = fileLength;
                _pInBase = pIn;
                _pOutBase = pOut;
                _numberStripes = numberStripes;
                _bufferSize = bufferSize;
                _optimize = optimize;
            }

            private int _threadNum;
            private Int64 _fileLength;
            private int _iShift;
            private byte* _pInBase;
            private byte* _pOutBase;
            private int _numberStripes;
            private int _bufferSize;
            private bool _optimize;

            public Int64 fileLength { get { return _fileLength; } }
            public int threadNumber { get { return _threadNum; } }
            public int iShift { get { return _iShift; } }
            public int numberStripes { get { return _numberStripes; } }
            public int bufferSize { get { return _bufferSize; } }
            public byte* pInBase { get { return _pInBase; } }
            public byte* pOutBase { get { return _pOutBase; } }
            public Int64 numBytesEncrypted { get { return _numBytesEncrypted; } set { _numBytesEncrypted = value; } }
            public bool optimize { get { return _optimize; } set { _optimize = value; } }
            public override String ToString()
            {
                return "InteropMemoryMappedCciImplementation.cciArgMM. thread/task: " + _threadNum.ToString() + ", Shift: " + _iShift.ToString() + ", Length: " + _fileLength.ToString() + ", Optimize: " + _optimize.ToString();
            }
        }

        /// <summary>
        /// Perform single threaded Mmeory Mapped conversion.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns>Number of bytes processed</returns>
        /// <remarks>This function contains both optimized and non-optimized code. The code to use is determined at run time based on the thread parameter.
        /// The optimized code is a "first attempt", and you can experiment with alternative implementations.</remarks>
        public Int64 cciMM(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            Int64 iPos = 0;
            byte* pIn = null, pOut = null;
            FileMm inMap = null, outMap = null;

            try
            {
                try
                {
                    FileInfo fi = new FileInfo(inFile);
                    Int64 fileSize = fi.Length;

                    inMap = new FileMm(inFile);
                    outMap = new FileMm(outFile, false, fileSize);
                    pIn = inMap.pFile;
                    pOut = outMap.pFile;

                    byte b;
                    if (!optimize)
                    {
                        for (iPos = 0; iPos < fileLength; iPos++)
                        {
                            b = *pIn;
                            *pOut = (byte)(b + iShift);
                            pIn++; pOut++;
                        }
                    }
                    else
                    {
                        for (iPos = 0; iPos < fileLength; iPos += 8)
                        {
                            pOut[iPos] = (byte)(pIn[iPos] + iShift);
                            pOut[iPos + 1] = (byte)(pIn[iPos + 1] + iShift);
                            pOut[iPos + 2] = (byte)(pIn[iPos + 2] + iShift);
                            pOut[iPos + 3] = (byte)(pIn[iPos + 3] + iShift);
                            pOut[iPos + 4] = (byte)(pIn[iPos + 4] + iShift);
                            pOut[iPos + 5] = (byte)(pIn[iPos + 5] + iShift);
                            pOut[iPos + 6] = (byte)(pIn[iPos + 6] + iShift);
                            pOut[iPos + 7] = (byte)(pIn[iPos + 7] + iShift);
                        }
                        pIn -= 8; pOut -= 8; iPos -= 8;
                        for (; iPos < fileLength; iPos++)
                        {
                            pOut[iPos] = (byte)(pIn[iPos] + iShift);
                        }
                    }
                }
                catch (Exception e)
                {
                    Console.Error.WriteLine("Error. cciMM. Bytes encrypted: {0}. {1}.", iPos, e.Message);
                    return iPos;
                }
            }
            finally
            {
                inMap.Dispose();
                outMap.Dispose();
            }
            return fileLength;
        }
    }
}

/// <summary>
/// ForAll extension that applies to all IEnumerable types
/// </summary>
/// <remarks>Copied from page 52 of "Effective C#: 50 Specific Ways to Improve Your C#" by Bill Wagner, Addison-Wesley Effective Software Development Series, 2010.
/// This is used in the multithreaded and thread pool implementations to help replace explicit loops with query syntax.</remarks>
public static class Extensions
{
    public static void ForAll<T>(
        this IEnumerable<T> sequence,
        Action<T> action)
    {
        foreach (T item in sequence)
            action(item);
    }
}