﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;
using System.Threading;
using System.Diagnostics;
#if MEMORY_MAPPED_FILES_DOTNET4
using System.IO.MemoryMappedFiles;
#endif

using FileMapped;

namespace cci
{
    class FileStreamCciImplementation
    {
        /// <summary>
        /// Delegate stripe conversion to numberStripes work items.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns>Number of bytes encrypted</returns>
        /// <remarks>See Duffy, "Concurrent Programming in Windows", pp 364-368 or Toub, "Patterns of Parallel Programming", pp. 8ff.</remarks>
        public Int64 cciFSTP(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            Int64 totalBytesEncrypted = 0;
            int remainingCallbacks = numberStripes;

            ManualResetEvent finishedEvent = new ManualResetEvent(false);

            // Create the conversion thread tasks to convert file stripes concurrently.
            cciArgFS[] threadArgs = (from n in Enumerable.Range(0, numberStripes) select new cciArgFS(inFile, outFile, iShift, bufferSize, numberStripes, n, optimize)).ToArray();

            // Create the pooled tasks and wait for them all to complete.
            threadArgs.ForAll((ta) =>
                ThreadPool.QueueUserWorkItem(delegate
                {
                    cciStripeFA(ta);
                    if (Interlocked.Decrement(ref remainingCallbacks) == 0)
                    {
                        finishedEvent.Set();
                    }
                })
            );

            // Wait for all pooled tasks to complete
            finishedEvent.WaitOne();

            // Accumulate the encrypted byte count and close the files by disposing of the arguments
            threadArgs.ForAll((ta) => { totalBytesEncrypted += ta.numBytesEncrypted; ta.Dispose(); });

            return totalBytesEncrypted;
        }

        /// <summary>
        /// Perform stripe conversion in numberStripes independent threads.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns></returns>
        public Int64 cciFSMT(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            Int64 totalBytesEncrypted = 0;

            // Create the conversion thread arguments, then create the threads to convert file stripes concurrently
            cciArgFS[] threadArgs = (from n in Enumerable.Range(0, numberStripes) select new cciArgFS(inFile, outFile, iShift, bufferSize, numberStripes, n, optimize)).ToArray();
            Thread[] cciThread = (from n in Enumerable.Range(0, numberStripes) select new Thread(new ParameterizedThreadStart(cciStripeFA))).ToArray();

            // Start all the threads. The static ForAll method is implemented at the end of the file.
            threadArgs.ForAll((ta) => cciThread[ta.threadNumber].Start(ta));

            // Join the threads, accumulate the byte count, and dispose of the thread argument
            // NOTE: Be sure to close the file streams in Dispose. I forgot to do this initially, and sometimes the next program could not open one of the
            // files as it was still in use (apparently, the garbage collector had not gotten around to closing the file).
            threadArgs.ForAll((ta) => { cciThread[ta.threadNumber].Join(); totalBytesEncrypted += ta.numBytesEncrypted; ta.Dispose(); });

            return totalBytesEncrypted;
        }

        /// <summary>
        /// cciStripeFA thread/work item function. FileStream version.
        /// </summary>
        /// <param name="cargObject">Contains all state for this thread, including FileStreams. Each thread requires separate FileStream state for the position.</param>
        /// <remarks>This function contains both optimized and non-optimized code. The code to use is determined at run time based on the thread parameter.
        /// The optimized code is a "first attempt", and you can experiment with alternative implementations.</remarks>
        private void cciStripeFA(object cciArgFSObject)
        {
            cciArgFS cciArgFS = cciArgFSObject as cciArgFS;
            if (cciArgFS == null)
            {
                // This should never happen.
                Console.Error.WriteLine("Error. cciStripeFA thread function. Argument object is no a cciArgFA.");
                return; // The error will be detected as a failure to convert all the bytes in the file.
            }
            // I prefer to move all the objecct parameters into local variables for safety. I've had problems when accessing the object in the conversion loop in the function body.
            int iThread = cciArgFS.threadNumber, bufferSize = cciArgFS.bufferSize,  iRead;
            byte iShift = (byte)cciArgFS.iShift;
            int numberStripes = cciArgFS.numberStripes;
            Int64 fileLength = cciArgFS.fileLength, numBytesEncrypted = cciArgFS.numBytesEncrypted;
            FileStream inStream = cciArgFS.inStream, outStream = cciArgFS.outStream;
            bool optimize = cciArgFS.optimize;

            byte[] record = new byte[bufferSize];
            Int64 curPosition;

            curPosition = bufferSize * iThread;
            while (curPosition < cciArgFS.fileLength)
            {
                inStream.Seek(curPosition, SeekOrigin.Begin);
                outStream.Seek(curPosition, SeekOrigin.Begin);
                iRead = inStream.Read(record, 0, bufferSize);
                byte bShift = (byte)iShift;
                if (!optimize)
                {
                    // Simple loop, one byte at a time. This is the default
                    for (int i = 0; i < iRead; i++)
                    {
                        record[i] += bShift;
                    }
                }
                else
                {
                    // Use any optimization technique you wish to evaluate.
                    // This implementation unwinds the loop, encypting 8 bytes at a time, followed by any residual bytes
                    int i = 0;
                    for (i = 0; i < iRead; i += 8)
                    {
                        record[i] += bShift;
                        record[i + 1] += bShift;
                        record[i + 2] += bShift;
                        record[i + 3] += bShift;
                        record[i + 4] += bShift;
                        record[i + 5] += bShift;
                        record[i + 6] += bShift;
                        record[i + 7] += bShift;
                    }
                    for (int j = i; j < iRead; j++) record[j] += bShift;
                }
                outStream.Write(record, 0, iRead);
                curPosition += (bufferSize * numberStripes);
                numBytesEncrypted += iRead;
            }
            cciArgFS.numBytesEncrypted = numBytesEncrypted;
        }

        /// <summary>
        /// Perform single threaded FileStream conversion.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <param name="optimize">Employ any code optimization technique you want to evaluate</param>
        /// <returns>Number of bytes processed</returns>
        /// <remarks>This function contains both optimized and non-optimized code. The code to use is determined at run time based on the thread parameter.
        /// The optimized code is a "first attempt", and you can experiment with alternative implementations.</remarks>
        public Int64 cciFS(String inFile, String outFile, int iShift, Int64 fileLength, int numberStripes, int bufferSize, bool optimize)
        {
            FileStream inStream = null, outStream = null;
            try
            {
                inStream = new FileStream(inFile, FileMode.Open, FileAccess.Read);
                outStream = new FileStream(outFile, FileMode.Create, FileAccess.ReadWrite);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error: Failed opening input file {0} or output file {1}. {2}", inFile, outFile, e.Message);
                return 0;
            }

            byte[] buffer = new byte[bufferSize]; // bufferSize is a global variable, set from the command line.
            Int64 filePos = 0;

            try
            {
                while (filePos < fileLength)
                {
                    int countRead = inStream.Read(buffer, 0, bufferSize);
                    byte bShift = (byte)iShift;
                    if (!optimize)
                    {
                        // Simple loop, one byte at a time. This is the default
                        for (int i = 0; i < countRead; i++)
                        {
                            buffer[i] += bShift;
                        }
                    }
                    else
                    {
                        // Use any optimization technique you wish to evaluate.
                        // This implementation unwinds the loop, encypting 8 bytes at a time, followed by any residual bytes
                        int i = 0;
                        for (i = 0; i < countRead; i += 8)
                        {
                            buffer[i] += bShift;
                            buffer[i + 1] += bShift;
                            buffer[i + 2] += bShift;
                            buffer[i + 3] += bShift;
                            buffer[i + 4] += bShift;
                            buffer[i + 5] += bShift;
                            buffer[i + 6] += bShift;
                            buffer[i + 7] += bShift;
                        }
                        for (int j = i; j < countRead; j++) buffer[j] += bShift;
                    }
                    outStream.Write(buffer, 0, countRead);
                    filePos += countRead;
                }
            }
            finally
            {
                inStream.Close(); outStream.Close();
            }
            return filePos;
        }

        /// <summary>
        /// Thread argument for the cci threads and work items. FileStream version.
        /// </summary>
        class cciArgFS : IDisposable
        {
            private Int64 _numBytesEncrypted = 0;

            public cciArgFS(String inFile, String outFile, int iShift, int bufferSize, int numberStripes, int iThread, bool optimize)
            {
                _threadNum = iThread;
                _iShift = iShift;
                _optimize = optimize;
                _bufferSize = bufferSize;
                _numberStripes = numberStripes;
                try
                {
                    _inStream = new FileStream(inFile, FileMode.Open, FileAccess.Read, FileShare.Read);
                    _outStream = new FileStream(outFile, FileMode.Create, FileAccess.ReadWrite, FileShare.ReadWrite);
                }
                catch (Exception e)
                {
                    Console.Error.WriteLine("Error: Failed opening input file {0} or output file {1}. {2}", inFile, outFile, e.Message);
                }
                _fileLength = inStream.Length;
            }

            public void Dispose()
            {
                _inStream.Close();
                _outStream.Close();
            }

            public override String ToString()
            {
                return "FileStreamCciImplementation.cciArgFS. thread/task: " + _threadNum.ToString() + ", Shift: " + _iShift.ToString() + ", Length: " + _fileLength.ToString() + ", Optimize: " + _optimize.ToString();
            }

            private FileStream _inStream;
            private FileStream _outStream;
            private int _threadNum;
            private Int64 _fileLength;
            private int _iShift;
            private bool _optimize;
            private int _bufferSize;
            private int _numberStripes;

            public FileStream inStream { get { return _inStream; } }
            public FileStream outStream { get { return _outStream; } }
            public Int64 fileLength { get { return _fileLength; } }
            public int threadNumber { get { return _threadNum; } }
            public int iShift { get { return _iShift; } }
            public int numberStripes { get { return _numberStripes; } }
            public int bufferSize { get { return _bufferSize; } }
            public Int64 numBytesEncrypted { get { return _numBytesEncrypted; } set { _numBytesEncrypted = value; } }
            public bool optimize { get { return _optimize; } set { _optimize = value; } }
        }
    }
}
