﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;
using System.Threading;
using System.Diagnostics;
#if MEMORY_MAPPED_FILES_DOTNET4
using System.IO.MemoryMappedFiles;
#endif

using FileMapped;
using FileMappedImplementations;

namespace cci
{
    /// <summary>
    /// Caesar Cipher test program with functions for combinations of FileStream, MemoryMapped, and threaded implementations.
    /// </summary>
    /// <remarks>The implementations illustrate alternative sequential file processing techniques. The code is monolithic in the sense that the actual implementation
    /// technique is decided at runtime on the basis of command line arguments, and the different implementations are all in this single source file. Furthermore,
    /// prototype optimized code is in the same functions as the pre-optimized code.
    /// TO DO: Assure all resource leaks are properly addressed and that exceptions and errors are properly handled.</remarks>
    class Program
    {
        private static int bufferSize = 4096;        // Default (typical page size): Global buffer size used wherever buffers are required. Used as the stripe size for MultiThreaded and ThreadPool concurrency.
        private static int numberStripes = Environment.ProcessorCount;       // Default: Maximum number of threads or work items running concurrently. No effect if threadingTechnique == "None"

        static void Main(string[] args)
        {
            String[] accessTechniques = { "FileStream", "MemoryMapped" };              // Default: FileStream
            String[] concurrencyTechniques = { "None", "NoneFast", "MultiThreaded", "MultiThreadedFast", "ThreadPool", "ThreadPoolFast" };   // Default: None

            if (args.Length < 3)
            {
                Console.Error.WriteLine("Usage Error: cci iShift inFile outFile [accessTechnique [ threadingTechnique [bufferSize [numberStripes ] ] ] ] ");
                return;
            }

            int iShift = int.Parse(args[0]);
            String inFile = args[1];
            String outFile = args[2];
            String accessTechnique = "FileStream";
            if (args.Length >= 4) accessTechnique = args[3];
            String threadingTechnique = "None";
            if (args.Length >= 5) threadingTechnique = args[4];
            if (args.Length >= 6) bufferSize = int.Parse(args[5]);
            if (args.Length >= 7) numberStripes = int.Parse(args[6]);

            if (!accessTechniques.Contains(accessTechnique))
            {
                Console.Error.WriteLine("Error: File access technique {0} is not supported.", accessTechnique);
                return;
            }
            if (!concurrencyTechniques.Contains(threadingTechnique))
            {
                Console.Error.WriteLine("Error: Threading technique {0} is not supported.", threadingTechnique);
                return;
            }

            Console.WriteLine("Encode file {0} with a shift of {1} to file {2}.", inFile, iShift, outFile);
            Console.WriteLine("The file access technique is: {0}. The threading technique is: {1}.", accessTechnique, threadingTechnique);
            Console.WriteLine("bufferSize: {0}, numberStripes: {1}.", bufferSize, numberStripes);
            Int64 startTime = DateTime.Now.Ticks, bytesEncrypted = 0, fileLength = 0;

            try
            {
                System.IO.FileInfo fileProps = new System.IO.FileInfo(inFile);
                fileLength = fileProps.Length;

                if (accessTechnique.Equals("FileStream"))
                {
                    FileStreamCciImplementation fsCciObject = new FileStreamCciImplementation();
                    if (threadingTechnique.Equals("None"))
                    {
                        bytesEncrypted = fsCciObject.cciFS(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("NoneFast"))
                    {
                        bytesEncrypted = fsCciObject.cciFS(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else if (threadingTechnique.Equals("MultiThreaded"))
                    {
                        bytesEncrypted = fsCciObject.cciFSMT(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("MultiThreadedFast"))
                    {
                        bytesEncrypted = fsCciObject.cciFSMT(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else if (threadingTechnique.Equals("ThreadPool"))
                    {
                        bytesEncrypted = fsCciObject.cciFSTP(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("ThreadPoolFast"))
                    {
                        bytesEncrypted = fsCciObject.cciFSTP(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else
                    {
                        Console.Error.WriteLine("Error. This access/threading combination is not supported.");
                    }
                }

                if (accessTechnique.Equals("MemoryMapped"))
                {
                    InteropMemoryMappedCciImplementation immCciObject = new InteropMemoryMappedCciImplementation();
                    if (threadingTechnique.Equals("None"))
                    {
                        bytesEncrypted = immCciObject.cciMM(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("NoneFast"))
                    {
                        bytesEncrypted = immCciObject.cciMM(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else if (threadingTechnique.Equals("MultiThreaded"))
                    {
                        bytesEncrypted = immCciObject.cciMMMT(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("MultiThreadedFast"))
                    {
                        bytesEncrypted = immCciObject.cciMMMT(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else if (threadingTechnique.Equals("ThreadPool"))
                    {
                        bytesEncrypted = immCciObject.cciMMTP(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, false);
                    }
                    else if (threadingTechnique.Equals("ThreadPoolFast"))
                    {
                        bytesEncrypted = immCciObject.cciMMTP(inFile, outFile, iShift, fileLength, numberStripes, bufferSize, true);
                    }
                    else
                    {
                        Console.Error.WriteLine("Error. This access/threading combination is not supported.");
                    }
                }
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error. Encoding failed. {0}", e.Message);
                return;
            }

            if (fileLength != bytesEncrypted)
            {
                Console.Error.WriteLine("Error. The input file was not completely processed. {0} bytes of {1} bytes were encrypted.", bytesEncrypted, fileLength);
                return;
            }

            Int64 elapsedTime = DateTime.Now.Ticks - startTime;

            Console.WriteLine("cci Program complete. Elapsed time: {0}.{1} seconds. Encrypted {2} bytes.", elapsedTime / 10000000, (elapsedTime % 10000000) / 10000, bytesEncrypted);
            Console.WriteLine("{0}, {1}, {2}, {3}, {4}.{5}, {6}", accessTechnique, threadingTechnique, bufferSize, numberStripes,
                elapsedTime / 10000000, (elapsedTime % 10000000) / 10000, bytesEncrypted);
            return;
        }

        // Some preliminary code using the .NET 4.0 MemoryMappedFile class. I was not able to get good performance as the
        // accessor does not provide a pointer; you still need to move data to and from a managed buffer.
#if MEMORY_MAPPED_FILES_DOTNET4
        /// <summary>
        /// Slow version of MM conversion. Several variations are listed (some are commented out)
        /// The problem is that you have to copy the input mapped memory into a buffer, modify the buffer, and
        /// then copy the buffer back to the output mapped buffer. Variations, such as using ReadByte to
        /// access one byte at a time are even slower. The solution is to use real pointers and the Windows
        /// API, as shown in cciMM.
        /// </summary>
        /// <param name="inFile"></param>
        /// <param name="outFile"></param>
        /// <param name="iShift"></param>
        /// <param name="fileLength"></param>
        /// <returns></returns>
        Int64 cciMM_Slow(String inFile, String outFile, int iShift, ref Int64 fileLength)
        {
            System.IO.FileInfo fileProps = new System.IO.FileInfo(inFile);
            fileLength = fileProps.Length;
            MemoryMappedFile inHandle = null, outHandle = null;
            try
            {
                inHandle = MemoryMappedFile.CreateFromFile(inFile, FileMode.Open, null, fileLength, MemoryMappedFileAccess.Read);
                outHandle = MemoryMappedFile.CreateFromFile(outFile, FileMode.Create, null, fileLength, MemoryMappedFileAccess.ReadWrite);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error: Failed creating mapping handle on input file {0} or output file {1}. {2}", inFile, outFile, e.Message);
                return 0;
            }

            MemoryMappedViewAccessor inAccessor = null, outAccessor = null;
            try
            {
                inAccessor = inHandle.CreateViewAccessor(0, fileLength, MemoryMappedFileAccess.Read);
                outAccessor = outHandle.CreateViewAccessor(0, fileLength, MemoryMappedFileAccess.ReadWrite);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error: Failed creating mapping view stream on input file {0} or output file {1}. {2}", inFile, outFile, e.Message);
                return 0;
            }

            //byte[] buffer = new byte[fileLength];
            Int64 iPosition = 0;
            try
            {
                byte[] buffer = new byte[fileLength];
                inAccessor.ReadArray<byte>(0, buffer, 0, (Int32)fileLength);

                for (iPosition = 0; iPosition < fileLength; iPosition++)
                {
                    buffer[iPosition] += (byte)iShift;
                }
                outAccessor.WriteArray<byte>(0, buffer, 0, (Int32)fileLength);
                //outAccessor.Write<LargeBuffer>(0, ref myBuffer);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error. cci_MM read/write loop at position {0}. {1}", iPosition, e.Message);
            }
            return iPosition;
        }
#endif
    }

    // A partial and incomplete attempt to use the Asynchronous Programming Model (APM). Completing this might be
    // interesting to someone. I will be surprised, however, if you get good performance. But, please feel free to prove me wrong!
#if USE_ASYNCH_IO
// This code is not complete. It attempts to use mulitple asynch operations, similar to cciOV and cciEX in WSP4,
// However, as far as I can tell, BeginRead and BeginWrite do not allow you to specify the file postion. You would
// need to use seek, which would mean you'd need multiple file streams, one for each stripe. This is not likely
// to work well, but maybe we can play with it later.
        /// <summary>
        /// Delegate stripe conversion to numberStripes asynchronous callbacks.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="outFile">Output file path name</param>
        /// <param name="iShift">Caesar cipher shift value; can be positive or negative.</param>
        /// <param name="fileLength">Input file length (Set by the method)</param>
        /// <returns>Number of bytes encrypted (Not meaningful in this implementation; it's always fileLength)</returns>
        /// <remarks>See Duffy, Concurrent Programming in Windows, pp 364-368.</remarks>
        Int64 cciFAAS(String inFile, String outFile, int iShift, Int64 fileLength)
        {
            FileStream inStream = null, outStream = null;
            try
            {
                inStream = new FileStream(inFile, FileMode.Open, FileAccess.Read, FileShare.Read, bufferSize, true);
                outStream = new FileStream(outFile, FileMode.Create, FileAccess.ReadWrite, FileShare.ReadWrite, bufferSize, true);
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error: Failed opening input file {0} or output file {1}. {2}", inFile, outFile, e.Message);
                return 0;
            }

            // Construct callback arguments for each callback, and initiate the first read operations for each stripe.
            cciArgFSAS[] callbackArgs = new cciArgFSAS[numberStripes];
            for (int iStripe = 0; iStripe < numberStripes; iStripe++)
            {
                callbackArgs[iStripe] = new cciArgFSAS(inStream, outStream, iShift, iStripe, bufferSize);
                inStream.BeginRead(callbackArgs[iStripe].buffer, 0, bufferSize, callbackFAAS, callbackArgs[iStripe]);
            }
            
            return filePos;
        }

        void callbackFAAS(IAsyncResult result)
        {
            cciArgFSAS arg = (cciArgFSAS)result;
            long fileLength = arg.fileLength;
            FileStream inStream = arg.inStream;
            FileStream outStream = arg.outStream;
            int stripeNum = arg.stripeNumber;
            int bufferSize = arg.bufferSize;
            int iShift = arg.iShift;
            byte[] buffer = arg.buffer;

            // Process the completed operation and begin the next operation (Write follows Read, and Read follows Write).

            if (arg.isRead)
            {
                int readCount = inStream.EndRead(result);
                
                for (int i = 0; i < readCount; i++)
                {
                    buffer[i] += (byte)iShift;
                }
                arg.isRead = false;
                outStream.BeginWrite(buffer, 0, bufferSize, callbackFAAS, arg);
                filePos += countRead;
            }
            else
            {
                int writeCount = outStream.EndWrite(result);
            }
            }

            inStream.Close(); outStream.Close();

        }

        class cciArgFSAS
        {
            public cciArgFSAS(FileStream inStream, FileStream outStream, int iShift, int iStripe, int bufferSize)
            {
                _stripeNum = iStripe;
                _iShift = iShift;
                _fileLength = inStream.Length;
                _inStream = inStream;
                _outStream = outStream;
                _bufferSize = bufferSize;
                _buffer = new byte[bufferSize];
                _isRead = true;                     // The first operation will be a read
            }

            private int _stripeNum;
            private Int64 _fileLength;
            private int _iShift;
            private int _bufferSize;
            private FileStream _inStream;
            private FileStream _outStream;
            private byte[] _buffer;
            private bool _isRead;

            public Int64 fileLength { get { return _fileLength; } }
            public int stripeNumber { get { return _stripeNum; } }
            public int iShift { get { return _iShift; } }
            public FileStream inStream { get { return _inStream; } }
            public FileStream outStream { get { return _outStream; } }
            public byte[] buffer { get { return _buffer; } }
            public int bufferSize { get { return _bufferSize; } }
            public bool isRead { get { return _isRead; } set { _isRead = value; } }
    }
#endif
}

